// Copyright 2020 the Pinniped contributors. All Rights Reserved.
// SPDX-License-Identifier: Apache-2.0

// Package symmetricsecrethelper provides a type that can generate and validate symmetric keys as
// Secret's.
package symmetricsecrethelper

import (
	"fmt"
	"io"

	corev1 "k8s.io/api/core/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/runtime/schema"

	configv1alpha1 "go.pinniped.dev/generated/1.19/apis/supervisor/config/v1alpha1"
	"go.pinniped.dev/internal/controller/supervisorconfig/generator"
)

const (
	// SymmetricSecretType is corev1.Secret.Type of all corev1.Secret's generated by this helper.
	SymmetricSecretType = "secrets.pinniped.dev/symmetric"
	// SymmetricSecretDataKey is the corev1.Secret.Data key for the symmetric key value generated by this helper.
	SymmetricSecretDataKey = "key"

	// symmetricKeySize is the default length, in bytes, of generated keys. It is set to 32 since this
	// seems like reasonable entropy for our keys, and a 32-byte key will allow for AES-256
	// to be used in our codecs (see dynamiccodec.Codec).
	symmetricKeySize = 32
)

type secretHelper struct {
	namePrefix string
	labels     map[string]string
	rand       io.Reader
	notifyFunc func(parent *configv1alpha1.OIDCProvider, child *corev1.Secret)
}

var _ generator.SecretHelper = &secretHelper{}

// New returns a SecretHelper that has been parameterized with common symmetric secret generation
// knobs.
func New(
	namePrefix string,
	labels map[string]string,
	rand io.Reader,
	notifyFunc func(parent *configv1alpha1.OIDCProvider, child *corev1.Secret),
) generator.SecretHelper {
	return &secretHelper{
		namePrefix: namePrefix,
		labels:     labels,
		rand:       rand,
		notifyFunc: notifyFunc,
	}
}

func (s *secretHelper) Name() string { return s.namePrefix }

// Generate implements SecretHelper.Generate().
func (s *secretHelper) Generate(parent *configv1alpha1.OIDCProvider) (*corev1.Secret, error) {
	key := make([]byte, symmetricKeySize)
	if _, err := s.rand.Read(key); err != nil {
		return nil, err
	}

	return &corev1.Secret{
		ObjectMeta: metav1.ObjectMeta{
			Name:      fmt.Sprintf("%s%s", s.namePrefix, parent.UID),
			Namespace: parent.Namespace,
			Labels:    s.labels,
			OwnerReferences: []metav1.OwnerReference{
				*metav1.NewControllerRef(parent, schema.GroupVersionKind{
					Group:   configv1alpha1.SchemeGroupVersion.Group,
					Version: configv1alpha1.SchemeGroupVersion.Version,
					Kind:    "OIDCProvider",
				}),
			},
		},
		Type: SymmetricSecretType,
		Data: map[string][]byte{
			SymmetricSecretDataKey: key,
		},
	}, nil
}

// IsValid implements SecretHelper.IsValid().
func (s *secretHelper) IsValid(parent *configv1alpha1.OIDCProvider, child *corev1.Secret) bool {
	if !metav1.IsControlledBy(child, parent) {
		return false
	}

	if child.Type != SymmetricSecretType {
		return false
	}

	key, ok := child.Data[SymmetricSecretDataKey]
	if !ok {
		return false
	}
	if len(key) != symmetricKeySize {
		return false
	}

	return true
}

// Notify implements SecretHelper.Notify().
func (s *secretHelper) Notify(parent *configv1alpha1.OIDCProvider, child *corev1.Secret) {
	s.notifyFunc(parent, child)
}
